from __future__ import annotations
from importlib.resources import files

import functools
import os
import re
from datetime import date, datetime
from pathlib import Path
from urllib.parse import quote

import pydantic as _p
from dotenv import load_dotenv
from fastapi.encoders import jsonable_encoder
from loguru import logger
from pydantic import Field, computed_field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from starlette.templating import Jinja2Templates

from shipaw.models.address import Address, Contact, FullContact
from shipaw.models.base import ShipawBaseModel
from shipaw.models.ship_types import ShipDirection


# def get_env2(env_name: str = 'SHIPAW_ENV') -> Path:
#     env = os.getenv(env_name)
#     if not env:
#         env_index = os.getenv('ENV_INDEX')
#         if env_index:
#             env_index_path = Path(env_index)
#             if env_index_path.exists():
#                 load_dotenv(env_index_path)
#                 env = os.getenv(env_name)
#                 if not env:
#                     raise ValueError(f'Environment variable {env_name} not set in {env_index_path}')
#             else:
#                 raise ValueError(f'ENV_INDEX ({env_index}) does not exist')
#         else:
#             raise ValueError(f'Environment variable {env_name} not set and ENV_INDEX not set')
#
#     env_path = Path(env)
#     if not env_path.exists():
#         raise ValueError(f'{env_path} not a valid path')
#     logger.debug(f'Loading environment from {env_path}')
#     return env_path

#
# def load_env_index(envs_index_key='SHIPAW_ENV_INDEX') -> None:
#     ei = os.environ.get(envs_index_key)
#     if not ei:
#         raise RuntimeError(f'Environment variable {envs_index_key} not set')
#     envs_index = Path(ei)
#     if not envs_index.exists():
#         raise FileNotFoundError(f'Environment index file {envs_index} does not exist')
#     logger.info(f'Loading env index from {envs_index}')
#     load_dotenv(envs_index)
#     for env in ('APC_ENV', 'PARCELFORCE_ENV', 'SHIPAW_ENV'):
#         if not os.getenv(env):
#             raise ValueError(f'Environment variable {env} not set in {envs_index}')
#         if not Path(os.getenv(env)).exists():
#             raise ValueError(f'Environment variable {env} points to non-existent file {os.getenv(env)}')
#

#
# def load_env() -> Path:
#     ei = Path(os.environ.get('SHIPAW_ENV_INDEX'))
#     logger.info(f'Loading env index from {ei}')
#     if not ei or not ei.exists():
#         raise ValueError(f'ENV_INDEX ({ei}) not set or does not exist')
#     load_env_index(ei)
#     shipaw_env = Path(os.getenv('SHIPAW_ENV'))
#     logger.debug(f'Loading SHIPAW environment from {shipaw_env}')
#     return shipaw_env


def get_shipaw_env(env_key:str = 'SHIPAW_ENV') -> Path:
    env = os.getenv(env_key)
    if not env:
        raise ValueError('SHIPAW_ENV not set')
    env_path = Path(env)
    if not env_path.exists():
        raise FileNotFoundError(f'SHIPAW_ENV file {env_path} does not exist')
    return env_path


def sanitise_id(value):
    return re.sub(r'\W|^(?=\d)', '_', value).lower()


def date_int_w_ordinal(n: int):
    """Convert an integer to its ordinal as a string, e.g. 1 -> 1st, 2 -> 2nd, etc."""
    return str(n) + ('th' if 4 <= n % 100 <= 20 else {1: 'st', 2: 'nd', 3: 'rd'}.get(n % 10, 'th'))


def ordinal_dt(dt: datetime | date) -> str:
    """Convert a datetime or date to a string with an ordinal day, e.g. 'Mon 1st Jan 2020'."""
    return dt.strftime(f'%a {date_int_w_ordinal(dt.day)} %b %Y')


def get_ui() -> Path:
    res = Path(files('shipaw'))
    res = res / 'ui'
    if not res.exists():
        raise FileNotFoundError(f'UI directory {res} does not exist')
    return res


class ProviderEnv(ShipawBaseModel):
    name: str
    env_file: Path


class ShipawSettings(BaseSettings):
    # toggles
    shipper_live: bool = False
    log_level: str = 'DEBUG'

    provider_dict: dict

    # dirs
    label_dir: Path
    log_dir: Path
    ui_dir: Path = Field(default_factory=get_ui)

    # auto dirs
    static_dir: Path | None = None
    template_dir: Path | None = None
    templates: Jinja2Templates | None = None

    # sender details
    address_line1: str
    address_line2: str | None = None
    address_line3: str | None = None
    town: str
    postcode: str
    country: str = 'GB'
    business_name: str
    contact_name: str
    email: str
    phone: str | None = None
    mobile_phone: str

    model_config = SettingsConfigDict()

    ## SET UI/TEMPLATE DIRS ##
    @model_validator(mode='after')
    def set_ui(self):
        self.static_dir = self.static_dir or self.ui_dir / 'static'
        self.template_dir = self.template_dir or self.ui_dir / 'templates'
        self.templates = self.templates or Jinja2Templates(directory=self.template_dir)
        self.templates.env.filters['jsonable'] = jsonable_encoder
        self.templates.env.filters['urlencode'] = lambda value: quote(str(value))
        self.templates.env.filters['sanitise_id'] = sanitise_id
        self.templates.env.filters['ordinal_dt'] = ordinal_dt
        return self

    ## SET LOGGING & LABELS ##
    @computed_field
    @property
    def log_file(self) -> Path:
        return self.log_dir / 'shipaw.log'

    @computed_field
    @property
    def ndjson_log_file(self) -> Path:
        return self.log_dir / 'shipaw.ndjson'

    @_p.model_validator(mode='after')
    def create_log_files(self):
        self.log_dir.mkdir(parents=True, exist_ok=True)
        for v in (self.log_file, self.ndjson_log_file):
            v.touch()
        return self

    @_p.field_validator('label_dir', mode='after')
    def create_label_dirs(cls, v, values):
        directions = [_ for _ in ShipDirection]
        for subdir in directions:
            apath = v / subdir
            if not apath.exists():
                apath.mkdir(parents=True, exist_ok=True)
        return v

    ## SET ADDRESS/CONTACT OBJECTS FROM ENV VARS ##
    @property
    def contact(self):
        return Contact(
            contact_name=self.contact_name,
            email_address=self.email,
            mobile_phone=self.mobile_phone,
        )

    @property
    def address(self):
        return Address(
            address_lines=[_ for _ in [self.address_line1, self.address_line2, self.address_line3] if _],
            town=self.town,
            postcode=self.postcode,
            country=self.country,
            business_name=self.business_name,
        )

    @property
    def full_contact(self) -> FullContact:
        return FullContact(
            address=self.address,
            contact=self.contact,
        )

    @classmethod
    @functools.lru_cache
    def from_env(cls, env_key='SHIPAW_ENV') -> ShipawSettings:
        return cls(_env_file=get_shipaw_env(env_key))


# SHIPAW_SETTINGS = ShipawSettings(_env_file=get_shipaw_env())
@functools.lru_cache
def shipaw_settings() -> ShipawSettings:
    return ShipawSettings.from_env()




import json
import time
from base64 import b64decode
from typing import ClassVar

import httpx
from pydantic import BaseModel
from apc_hypaship.config import APCSettings
from apc_hypaship.models.request.shipment import Shipment as ShipmentAPC

from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.services import Services
from shipaw.models.shipment import Shipment, Shipment as ShipmentAgnost
from shipaw.providers.apc.response import booking_has_errors, errored_booking
from shipaw.providers.providers import ShippingProvider, register_provider
from shipaw.providers.apc.provider_funcs import (
    APC_SERVICES,
    apc_shipment_from_agnostic,
    apc_shipment_to_agnostic,
)


# @dataclass
@register_provider
class APCShippingProvider(ShippingProvider):
    name = 'APC'
    services: ClassVar[Services] = APC_SERVICES
    settings_type: ClassVar[APCSettings] = APCSettings

    def agnostic_shipment(self, shipment: ShipmentAPC) -> Shipment:
        return apc_shipment_to_agnostic(shipment)

    def provider_shipment(self, shipment: Shipment) -> BaseModel:
        return apc_shipment_from_agnostic(shipment)

    def book_shipment(self, shipment: dict | ShipmentAgnost) -> ShipmentBookingResponse:
        """Takes provider ShipmnentDict, or ShipmentAgnost object"""
        shipment_dict = self.build_request_json(shipment)
        res_json = self.get_response_json(shipment_dict)

        if booking_has_errors(shipment_dict):
            return errored_booking(shipment, res_json)

        resp = self.build_response(res_json, shipment)
        resp.label_data = self.wait_label(resp.shipment_num)
        return resp

    @staticmethod
    def build_response(res_json, shipment):

        orders = res_json['Orders']
        order = orders['Order']
        return ShipmentBookingResponse(
            shipment=shipment,
            shipment_num=(order['OrderNumber']),
            tracking_link='NOT IMPLEMENTED',
            data=res_json,
            status=(str(orders.get('Messages').get('Code'))),
            success=(orders.get('Messages').get('Code') == 'SUCCESS'),
        )

    def get_response_json(self, shipment_dict):
        res = httpx.post(self.settings.orders_endpoint, headers=self.settings.headers, json=shipment_dict)
        res.raise_for_status()
        return res.json()

    def build_request_json(self, shipment):
        apc_shipment = self.provider_shipment(shipment)
        shipment_dict = apc_shipment.model_dump(mode='json', by_alias=True)
        return shipment_dict

    def get_label_content(self, shipment_num: str) -> bytes:
        params = {'labelformat': 'PDF'}
        settings = self.settings
        label = httpx.get(settings.one_order_endpoint(shipment_num), headers=settings.headers, params=params)
        label.raise_for_status()
        content = label.json()['Orders']['Order']['Label']['Content']
        return b64decode(content)




from pathlib import Path

from combadge.core.errors import BackendError
from fastapi import APIRouter, Body, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.staticfiles import StaticFiles

from parcelforce_expresslink.address import AddressChoice as AddressChoicePF, Contact as ContactPF
from parcelforce_expresslink.client import ParcelforceClient
from shipaw.config import shipaw_settings, ShipawSettings
from shipaw.fapi.alerts import Alert, AlertType, Alerts, maybe_alert_phone_number
from shipaw.fapi.backend import get_el_client, try_book_shipment
from shipaw.fapi.form_data import shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import AddressRequest, ShipmentRequest
from shipaw.fapi.responses import ShipawTemplate, ShipawTemplateResponse
from shipaw.models.address import Address, AddressChoice as AddressChoiceAgnost
from shipaw.models.logging import log_obj
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment
from shipaw.providers.parcelforce.provider_funcs import (
    address_from_agnostic,
    full_contact_from_provider_contact_address,
)

router = APIRouter()
router.mount('/static', StaticFiles(directory=str(ShipawSettings.from_env().static_dir)), name='static')


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def ship_form(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj(shipment, 'Shipment received at /ship_form:')
    alerts: Alerts = request.app.alerts

    if any(['prdev' in str(_).lower() for _ in Path(__file__).parents]):
        msg = '"prdev" in cwd tree - BETA MODE - This is a development version'
        logger.warning(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)

    if shipaw_settings().shipper_live:
        msg = 'Shipper Live is True - Real Shipments will be booked'
    else:
        msg = 'Shipper_live is False - No Shipments will be booked'
    logger.warning(msg)
    alerts += Alert(message=msg, type=AlertType.NOTIFICATION)

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    alerts += await maybe_alert_apc(shipment_request)

    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


async def maybe_alert_apc(shipment_request):
    alerts = Alerts.empty()
    if shipment_request.provider_name == 'APC' and shipment_request.shipment.direction == ShipDirection.DROPOFF:
        alerts += Alert(
            message='APC does not support drop-off shipments - please select Outbound or Inbound Collection',
            type=AlertType.ERROR,
        )
    return alerts


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response = await try_book_shipment(shipment_request)
    await shipment_request.provider.handle_response_async(shipment_request, shipment_response)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.post('/addr_choices', response_model=list[AddressChoiceAgnost], response_class=JSONResponse)
async def get_addr_choices(
    request: Request,
    body: AddressRequest = Body(...),
    el_client: ParcelforceClient = Depends(get_el_client),
) -> list[AddressChoiceAgnost]:
    """Fetch candidate address choices for a postcode, optionally scored by closeness to provided address.

    Args:
        request: Request - FastAPI request object
        body: Address - request body containing postcode and optional address
        el_client: ELClient - Parcelforce ExpressLink client
    """
    postcode = body.postcode
    address_agnost = body.address
    pf_address = address_from_agnostic(address_agnost) if address_agnost else None
    # log_obj(pf_address, 'Address received at /cand:')

    try:
        res = el_client.get_choices(postcode=postcode, address=pf_address)
        res_agnost = [await convert_choice(_) for _ in res]
        # log_objs(res_agnost, 'Address choices returned from /cand:')
        return res_agnost

    except BackendError as e:
        alert = Alert(
            message=f'Error fetching candidates: {e}',
            type=AlertType.ERROR,
        )
        request.app.alerts += alert
        logger.warning(f'Error fetching candidates: {e}')
        addr = Address(address_lines=['ERROR:', str(e)], town='Error', postcode='Error', business_name='Error')
        chc = AddressChoiceAgnost(address=addr, score=0)
        return [chc]


async def convert_choice(choice: AddressChoicePF) -> AddressChoiceAgnost:
    fc = full_contact_from_provider_contact_address(contact=ContactPF.empty(), address=choice.address)
    return AddressChoiceAgnost(address=fc.address, score=choice.score)





from shipaw.providers.parcelforce.provider import ParcelforceShippingProvider
from shipaw.providers.providers import ShippingProvider
from shipaw.providers.apc.provider import APCShippingProvider

PROVIDER_TYPE_REGISTER: dict[str, type[ShippingProvider]] = {
    'APC': APCShippingProvider,
    'PARCELFORCE': ParcelforceShippingProvider,
}


from shipaw.fapi.alerts import Alerts, Alert
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.shipment import Shipment as ShipmentAgnost


def booking_has_errors(res_json: dict):
    if messages := res_json.get('Orders').get('Order').get('Messages'):
        return 'ErrorFields' in messages.keys()
    return False


def errored_booking(shipment: ShipmentAgnost, res_json: dict):
    messages = res_json.get('Orders').get('Order').get('Messages')
    fieldname, message = strip_apc_error_msgs(messages)
    return errored_deets(fieldname, message, shipment, res_json)


def strip_apc_error_msgs(messages):
    fieldname = messages['ErrorFields']['ErrorField']['FieldName']
    message = messages['ErrorFields']['ErrorField']['ErrorMessage']
    return fieldname, message


def errored_deets(fieldname: str, message: str, shipment: ShipmentAgnost, data: dict):
    alerts = Alerts(alert=[Alert(message=f'Error booking shipment: {fieldname}: {message}')])
    return ShipmentBookingResponse(
        alerts=alerts,
        shipment=shipment,
        shipment_num='FAILED TO BOOK',
        tracking_link='NOT IMPLEMENTED',
        data=data,
        success=False,
    )

